﻿// 
// Copyright (c) Microsoft. All rights reserved.
// Licensed under the MIT license.
// 
// Microsoft Bot Framework: http://botframework.com
// 
// Bot Builder SDK Github:
// https://github.com/Microsoft/BotBuilder
// 
// Copyright (c) Microsoft Corporation
// All rights reserved.
// 
// MIT License:
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED ""AS IS"", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
// NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
// OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
// WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

using System;
using System.Collections.Generic;
using System.Linq;

namespace Microsoft.Bot.Builder.FormFlow.Advanced
{
    /// <summary>
    /// Confirmation 
    /// </summary>
    /// <typeparam name="T">Form state.</typeparam>
    public class Confirmation<T> : Field<T>
        where T : class
    {
        /// <summary>
        /// Construct a confirmation.
        /// </summary>
        /// <param name="prompt">Confirmation prompt expressed using \ref patterns.</param>
        /// <param name="condition">Delegate for whether confirmation applies.</param>
        /// <param name="dependencies">Fields that must have values before confirmation can run.</param>
        public Confirmation(PromptAttribute prompt, ConditionalDelegate<T> condition, IEnumerable<string> dependencies)
            : base(Guid.NewGuid().ToString(), FieldRole.Confirm)
        {
            SetPrompt(prompt);
            _condition = condition;
            _dependencies = dependencies.ToArray();
            var noStep = (dependencies.Count() > 0 ? new NextStep(dependencies) : new NextStep());
            _next = (value, state) => value ? new NextStep() : noStep;
        }

        public override object GetValue(T state)
        {
            return null;
        }

        public override IEnumerable<string> Dependencies
        {
            get
            {
                return _dependencies;
            }
        }

        #region IFieldPrompt
        public override bool Active(T state)
        {
            return _condition(state);
        }

        public override NextStep Next(object value, T state)
        {
            return _next((bool) value, state);
        }

        public override void SetValue(T state, object value)
        {
            throw new NotImplementedException();
        }

        public override bool IsUnknown(T state)
        {
            return true;
        }

        public override void SetUnknown(T state)
        {
            throw new NotImplementedException();
        }

        public override IPrompt<T> Prompt()
        {
            if (_prompter == null)
            {
                _prompter = new Prompter<T>(_promptDefinition, _form, new RecognizeBool<T>(this));
            }
            return _prompter;
        }
        #endregion

        #region Implementation
        private delegate NextStep NextDelegate(bool response, T state);
        private readonly ConditionalDelegate<T> _condition;
        private readonly string[] _dependencies;
        private IPrompt<T> _prompter;
        private readonly NextDelegate _next;
        #endregion
    }
}
